	/**
	 *	@brief decompression routine
	 *
	 *	@param[in] p_s_base64 is null-terminated string containing Base-64 encoded compressed data
	 *	@param[in] n_data_size is size of the decompressed data
	 *	@param[in] n_huff_stage_size is size of the compressed data before Huffman encoding
	 *	@param[in] n_bwt_primary_index is primary index for the BWT algorithm
	 *
	 *	@return Returns pointer to the decompressed data on success, 0 on failure.
	 *
	 *	@note The data (although it may be binary), is additionally null terminated.
	 */
	static uint8_t *p_Decode_BWT_MTF_RLE_Huff_Base64(const char *p_s_base64,
		size_t n_data_size, size_t n_huff_stage_size, uint32_t n_bwt_primary_index)
	{
		size_t n_base64_length = strlen(p_s_base64);
		size_t n_base64_out_size = (n_base64_length / 4) * 3 +
			max(n_base64_length % 4, size_t(1)) - 1;
		size_t n_prehuff_size = max(n_base64_out_size, n_data_size);
		size_t n_temp_buffer_size = max(n_prehuff_size + n_huff_stage_size,
			n_data_size * (1 + sizeof(uint32_t)));

		uint8_t *p_temp_buffer, *p_data;
		if(!(p_temp_buffer = new(std::nothrow) uint8_t[n_temp_buffer_size]) ||
		   !(p_data = new(std::nothrow) uint8_t[n_data_size + 1])) {
			if(p_temp_buffer)
				delete[] p_temp_buffer;
			return 0;
		}

		p_data[n_data_size] = 0; // !!
		{
			int p_ic_tab[128] = {0};
			for(int i = 0, j = 65, k = -62; i < 64; ++ i, ++ j, ++ k)
				p_ic_tab[(i < 52)? j + i / 26 * 6 : (k < 0)? i - 4 : "+/"[k]] = i;
			const uint8_t *p_src = (uint8_t*)p_s_base64;
			uint8_t *p_dst = p_temp_buffer;
			for(const uint8_t *p_end = p_src + n_base64_length - 3; p_src < p_end; p_src += 4, p_dst += 3) {
				p_dst[0] = (p_ic_tab[p_src[0]] << 2) | (p_ic_tab[p_src[1]] >> 4);
				p_dst[1] = (p_ic_tab[p_src[1]] << 4) | (p_ic_tab[p_src[2]] >> 2);
				p_dst[2] = (p_ic_tab[p_src[2]] << 6) | (p_ic_tab[p_src[3]]);
			}
			if(n_base64_length % 4 == 3) {
				p_dst[0] = (p_ic_tab[p_src[0]] << 2) | (p_ic_tab[p_src[1]] >> 4);
				p_dst[1] = (p_ic_tab[p_src[1]] << 4) | (p_ic_tab[p_src[2]] >> 2); // fixed byte order
			} else if(n_base64_length % 4 == 2)
				p_dst[0] = (p_ic_tab[p_src[0]] << 2) | (p_ic_tab[p_src[1]] >> 4);
		}
		{
			const uint8_t *p_huff_src = p_temp_buffer, *p_huff_end = p_huff_src + n_base64_out_size;
			_ASSERTE(p_huff_src + 16 * sizeof(int32_t) <= p_huff_end);
			const int32_t *p_code_num = (const int32_t*)p_huff_src;
			p_huff_src += 16 * sizeof(int32_t);
			int p_max_code[16], p_table_off[16], n_code_num = 0;
			for(int i = 0, n_code_word = 0; i < 16; ++ i, n_code_word <<= 1) {
				p_table_off[i] = (n_code_num += p_code_num[i]) - (n_code_word += p_code_num[i]);
				p_max_code[i] = n_code_word;
			}
			_ASSERTE(p_huff_src + n_code_num <= p_huff_end);
			const uint8_t *p_symbol = p_huff_src;
			p_huff_src += n_code_num - 1;
			uint8_t *p_dest = p_temp_buffer + n_prehuff_size, *p_end = p_dest + n_huff_stage_size;
			for(uint8_t n_byte = 0, n_bit_num = 7; p_dest != p_end; ++ p_dest, n_byte <<= 1) {
				for(int i = 0, n_code = 0;; ++ i, n_byte <<= 1) {
					_ASSERTE(i < 16);
					if(!((++ n_bit_num) & 7)) { // every 8
						_ASSERTE(p_huff_src < p_huff_end);
						n_byte = *(++ p_huff_src);
					}
					n_code = (n_code << 1) | (n_byte >> 7);
					if(n_code < p_max_code[i]) {
						_ASSERTE(n_code >= p_max_code[i] - p_code_num[i]); // n_code >= p_min_code[i]
						*p_dest = p_symbol[n_code + p_table_off[i]];
						break;
					}
				}
			}
		}
		{
			const uint8_t *p_rle_src = p_temp_buffer + n_prehuff_size,
				*p_src_end = p_rle_src + n_huff_stage_size;
			for(uint8_t *p_rle_dest = p_temp_buffer; p_rle_src != p_src_end;) {
				uint8_t n_code = *p_rle_src;
				bool b_compressed = n_code & 1;
				int n_run_length = (n_code >> 1) + 2;
				_ASSERTE(p_rle_src + ((b_compressed)? 2 : n_run_length) <= p_src_end);
				for(; -- n_run_length; ++ p_rle_dest)
					*p_rle_dest = (b_compressed)? p_rle_src[1] : *(++ p_rle_src);
				p_rle_src += int(1 << ((b_compressed)? 1 : 0)); // fix an x64 warning here
			}
		}
		{
			uint8_t p_buckets[256] = {0}; // each bucket value encoded as offset to its index
			uint8_t *p_data = p_temp_buffer, *p_end = p_temp_buffer + n_data_size;
			for(; p_data != p_end; ++ p_data) {
				uint8_t n_index = *p_data;
				*p_data = p_buckets[n_index] + n_index;
				if(n_index) {
					for(int i = n_index; i > 0; -- i)
						p_buckets[i] = p_buckets[i - 1] - 1;
					p_buckets[0] = *p_data;
				}
			}
		}
		{
			uint32_t p_buckets[256] = {0}, n_sum = 0;
			uint32_t *p_index_list = (uint32_t*)(p_temp_buffer + n_data_size);
			const uint8_t *p_src = p_temp_buffer, *p_end = p_temp_buffer + n_data_size;
			for(uint32_t *p_dest_idx = p_index_list; p_src != p_end; ++ p_src, ++ p_dest_idx)
				*p_dest_idx = (++ p_buckets[*p_src]) - 1;
 			for(int i = 0; i < 256; ++ i) {
				uint32_t n_value = p_buckets[i];
				p_buckets[i] = n_sum;
				n_sum += n_value;
			}
			uint8_t *p_dest = p_data + n_data_size - 1, *p_begin_1 = p_data - 1;
			for(uint32_t i = n_bwt_primary_index; p_dest != p_begin_1; -- p_dest,
			   i = p_index_list[i] + p_buckets[p_temp_buffer[i]])
				*p_dest = p_temp_buffer[i];
		}
		_ASSERTE(!p_data[n_data_size]); // make sure it was not overwritten

		delete[] p_temp_buffer;
		
		return p_data;
	}
